from typing import List
from pyspark.sql import DataFrame
from pyspark.sql.window import Window
from pyspark.sql.functions import row_number, col, lit, coalesce


def merge_and_retain_last(
    df1: DataFrame,
    df2: DataFrame,
    key_fields: List[str] = [],
    sort_field: str = None,
    deleted_field: str = None,
) -> DataFrame:
    """
    Merges the provided data frames together and returns a new DataFrame that contains only the most recent record
    for each value within the given `key_fields`

    key_fields : List[str]
        A list of fields that uniquely identify the entity being merged
    sort_field : str
        A field (usually a timestamp) to use for determining what record to keep. The record with the greatest value is retained.
    deleted_field : str
        The field to use as a deleted indicator. If this is True then the row is considered a delete marker.
    """
    window = Window.partitionBy(key_fields).orderBy(col(sort_field).desc())

    def add_deleted_flag_if_needed(df: DataFrame) -> DataFrame:
        if deleted_field not in df.columns:
            return df.withColumn(deleted_field, lit(False))
        return df

    row_num_col = "__row_num"

    df1 = add_deleted_flag_if_needed(df1)
    df2 = add_deleted_flag_if_needed(df2)
    merged = (
        df1.unionByName(df2)
        .select("*", row_number().over(window).alias(row_num_col))
        .filter(col(row_num_col) == 1)
        .filter(coalesce(deleted_field, lit(False)) == False)
        .drop(row_num_col)
        .drop(deleted_field)
    )
    return merged



import os
import re
import boto3
from typing import List
from cfn_tools import load_yaml
from awsglue import DynamicFrame
from awsglue.context import GlueContext
from pyspark.sql import SparkSession, DataFrame
from pyspark.sql.types import (
    StructType,
    StructField,
    StringType,
    ShortType,
    ByteType,
    IntegerType,
    LongType,
    DateType,
    TimestampType,
    DecimalType,
    ArrayType,
)
from functools import wraps
from typing import Callable


def coerce_to_schema(df: DataFrame, schema: StructType):
    # removes extraneous fields
    result = df.select(schema.fieldNames())

    for field in schema.fields:
        result = result.withColumn(field.name, result[field.name].cast(field.dataType))

    return result


def __get_glue_table(resources: dict, table_name: str):
    for _logical_id, resource_definition in resources.items():
        try:
            name = resource_definition["Properties"]["TableInput"]["Name"]
            if resource_definition["Type"] == "AWS::Glue::Table" and name == table_name:
                return resource_definition
        except KeyError:
            pass


class BadTypeException(Exception):
    pass


def __handle_decimal(match: re.Match):
    (precision, scale) = match.groups()
    return DecimalType(int(precision), int(scale))


def __handle_hive_struct_field(struct_field: str):
    [name, field_type] = struct_field.split(":")
    return StructField(name, __convert_type(field_type))


def __handle_hive_struct(hive_struct_types: str):
    return [__handle_hive_struct_field(field) for field in hive_struct_types.split(",")]


def __handle_struct(match: re.Match):
    (struct_fields,) = match.groups()
    return StructType(__handle_hive_struct(struct_fields))


def __handle_array(match: re.Match):
    (array_type,) = match.groups()
    return ArrayType(__convert_type(array_type))


def __convert_type(column_type: str):
    options = [
        (lambda t: t == "string", lambda _match: StringType()),
        (lambda t: t == "tinyint", lambda _match: ByteType()),
        (lambda t: t == "smallint", lambda _match: ShortType()),
        (lambda t: t == "int", lambda _match: IntegerType()),
        (lambda t: t == "bigint", lambda _match: LongType()),
        (lambda t: t == "date", lambda _match: DateType()),
        (lambda t: t == "timestamp", lambda _match: TimestampType()),
        (lambda t: re.match(r"decimal\((\d+),(\d+)\)", t), __handle_decimal),
        (lambda t: re.match(r"^struct<(.*)>$", t), __handle_struct),
        (lambda t: re.match(r"^array<(.*)>$", t), __handle_array),
    ]
    for test_func, type_func in options:
        match = test_func(column_type.lower())
        if match:
            return type_func(match)
    raise BadTypeException(f"could not find type for {column_type}")


def __convert_column(column):
    return StructField(column["Name"], __convert_type(column["Type"]))


def __convert_columns(columns: List[dict]):
    return [__convert_column(column) for column in columns]


def schema_from_cloudformation(path_to_template: str, table_name: str) -> StructType:
    template_text = open(path_to_template, "r").read()
    parsed_template = load_yaml(template_text)
    resources = parsed_template["Resources"]
    table = __get_glue_table(resources, table_name)
    return StructType(
        __convert_columns(
            table["Properties"]["TableInput"]["StorageDescriptor"]["Columns"]
        )
    )


def schema_from_glue(database_name: str, table_name: str):
    glue = boto3.client("glue", os.environ["AWS_REGION"])
    res = glue.get_table(DatabaseName=database_name, Name=table_name)
    return StructType(__convert_columns(res["Table"]["StorageDescriptor"]["Columns"]))


def __coalesce(*args):
    for arg in args:
        if arg is not None:
            return arg


def __get_schema_args(*args):
    first_non_none = __coalesce(*args)
    if callable(first_non_none):
        return first_non_none()
    return first_non_none


def __force_frame_to_fit(schema, *args, **kwargs):
    new_args = [*args]
    for i in range(len(args)):
        if isinstance(new_args[i], DynamicFrame):
            new_args[i] = new_args[i].toDF()
        if isinstance(new_args[i], DataFrame):
            new_args[i] = coerce_to_schema(new_args[i], schema)

    new_kwargs = {**kwargs}
    for k, _ in new_kwargs.items():
        if isinstance(new_kwargs[k], DynamicFrame):
            new_kwargs[k] = new_kwargs[k].toDF()
        if isinstance(new_kwargs[k], DataFrame):
            new_kwargs[k] = coerce_to_schema(new_kwargs[k], schema)
    return new_args, new_kwargs


def schema(*, schema_obj=None, schema_func=None):
    """
    Takes a pyspark schema or a function resulting in a pyspark schema and coerces the input frame to match this schema.
    """

    def annotation_func(func):
        @wraps(func)
        def wrapper_func(*args, **kwargs):
            final_schema = __get_schema_args(schema_obj, schema_func)
            new_args, new_kwargs = __force_frame_to_fit(final_schema, *args, **kwargs)
            return func(*new_args, **new_kwargs)

        return wrapper_func

    return annotation_func


def source(database: str, table: str, schema_obj=None, schema_func=None):
    spark = SparkSession.builder.getOrCreate()
    glue_context = GlueContext(spark.sparkContext)

    def annotation_func(func: Callable):
        @wraps(func)
        def wrapper_func(*args, **kwargs):
            final_schema = __get_schema_args(schema_obj, schema_func)
            dyn_frame = glue_context.create_dynamic_frame_from_catalog(
                database=database,
                table_name=table,
                transformation_ctx=f"source-{database}-{table}",
            )
            df = dyn_frame.toDF()
            fitted_frame = coerce_to_schema(df, final_schema)
            # Do we want to assume this is right here,
            # or do we want to do crazy reflection magic
            # to figure out a keyword arg? Both?
            return func(fitted_frame, *kwargs, **kwargs)

        return wrapper_func

    return annotation_func


def sink(database: str, table: str, schema_obj=None, schema_func=None):
    spark = SparkSession.builder.getOrCreate()
    glue_context = GlueContext(spark.sparkContext)

    def annotation_func(func: Callable):
        @wraps(func)
        def wrapper_func(*args, **kwargs):
            df = func(*args, **kwargs)
            final_schema = __get_schema_args(schema_obj, schema_func)
            fitted_frame = coerce_to_schema(df, final_schema)
            dyf = DynamicFrame.fromDF(
                fitted_frame, glue_context, f"sink-{database}-{table}"
            )
            glue_context.write_dynamic_frame_from_catalog(dyf, database, table)
            return fitted_frame

        return wrapper_func

    return annotation_func



import pytest
from pyspark import SparkContext
from pyspark.sql import SparkSession


@pytest.fixture(scope="session")
def spark_context():
    # TODO: move me when the fixtures is merged
    spark = SparkSession.builder.master("local[1]").getOrCreate()
    yield spark.sparkContext


@pytest.fixture()
def spark_session(spark_context: SparkContext):
    yield SparkSession(spark_context)



from datetime import datetime
import pytest
from glue_biscuit.test import DataFrameMatcher
from pyspark.sql import SparkSession
from pyspark.sql.types import (
    StructType,
    StructField,
    IntegerType,
    TimestampType,
    BooleanType,
    StringType,
)
from glue_biscuit.merge_utils import merge_and_retain_last


@pytest.fixture
def record_schema():
    return StructType(
        [
            StructField("id", IntegerType(), False),
            StructField("value", StringType(), False),
            StructField("_as_of", TimestampType(), False),
            StructField("_deleted", BooleanType(), True),
        ]
    )


def test_merge_empties(spark_session: SparkSession, record_schema: StructType):
    a = spark_session.createDataFrame([], record_schema)
    b = spark_session.createDataFrame([], record_schema)
    merged = merge_and_retain_last(
        a, b, key_fields=["id"], sort_field="_as_of", deleted_field="_deleted"
    )

    assert 0 == merged.count()


def test_merge_single_to_empty(spark_session: SparkSession, record_schema: StructType):
    record = {
        "id": 1,
        "value": "a",
        "_as_of": datetime.fromisoformat("2020-01-01T10:00:00"),
    }
    a = spark_session.createDataFrame([record], record_schema)
    b = spark_session.createDataFrame([], record_schema)
    merged = merge_and_retain_last(
        a, b, key_fields=["id"], sort_field="_as_of", deleted_field="_deleted"
    )
    merged.show()
    assert DataFrameMatcher([record]) == merged


def test_empty_to_single(spark_session: SparkSession, record_schema: StructType):
    record = {
        "id": 1,
        "value": "a",
        "_as_of": datetime.fromisoformat("2020-01-01T10:00:00"),
    }
    a = spark_session.createDataFrame([], record_schema)
    b = spark_session.createDataFrame([record], record_schema)
    merged = merge_and_retain_last(
        a, b, key_fields=["id"], sort_field="_as_of", deleted_field="_deleted"
    )

    assert merged == DataFrameMatcher([record])


def test_same_record_in_both_sets(
    spark_session: SparkSession, record_schema: StructType
):
    record = {
        "id": 1,
        "value": "a",
        "_as_of": datetime.fromisoformat("2020-01-01T10:00:00"),
    }
    a = spark_session.createDataFrame([record], record_schema)
    b = spark_session.createDataFrame([record], record_schema)
    merged = merge_and_retain_last(
        a, b, key_fields=["id"], sort_field="_as_of", deleted_field="_deleted"
    )

    assert merged == DataFrameMatcher([record])


def test_record_overwritten(spark_session: SparkSession, record_schema: StructType):
    record_a = {
        "id": 1,
        "value": "b",
        "_as_of": datetime.fromisoformat("2020-01-01T10:00:01"),
    }
    record_b = {
        "id": 1,
        "value": "b",
        "_as_of": datetime.fromisoformat("2020-01-01T10:00:00"),
    }
    merged = merge_and_retain_last(
        spark_session.createDataFrame([record_a], record_schema),
        spark_session.createDataFrame([record_b], record_schema),
        key_fields=["id"],
        sort_field="_as_of",
        deleted_field="_deleted",
    )

    assert merged == DataFrameMatcher([record_a])


def test_old_record_passed_in(spark_session: SparkSession, record_schema: StructType):
    record_a = {
        "id": 1,
        "value": "a",
        "_as_of": datetime.fromisoformat("2020-01-01T09:40:00"),
    }
    record_b = {
        "id": 1,
        "value": "b",
        "_as_of": datetime.fromisoformat("2020-01-01T10:00:00"),
    }
    merged = merge_and_retain_last(
        spark_session.createDataFrame([record_a], record_schema),
        spark_session.createDataFrame([record_b], record_schema),
        key_fields=["id"],
        sort_field="_as_of",
        deleted_field="_deleted",
    )

    assert merged == DataFrameMatcher([record_b])


def test_record_deleted(spark_session: SparkSession, record_schema: StructType):
    record_a = {
        "id": 1,
        "value": "b",
        "_as_of": datetime.fromisoformat("2020-01-01T10:00:01"),
        "_deleted": True,
    }
    record_b = {
        "id": 1,
        "value": "b",
        "_as_of": datetime.fromisoformat("2020-01-01T10:00:00"),
    }
    merged = merge_and_retain_last(
        spark_session.createDataFrame([record_a], record_schema),
        spark_session.createDataFrame([record_b], record_schema),
        key_fields=["id"],
        sort_field="_as_of",
        deleted_field="_deleted",
    )

    assert merged.count() == 0


def test_correct_record_is_updated(
    spark_session: SparkSession, record_schema: StructType
):
    record_a1 = {
        "id": 1,
        "value": "a",
        "_as_of": datetime.fromisoformat("2020-01-01T10:00:01"),
    }
    record_a2 = {
        "id": 2,
        "value": "a",
        "_as_of": datetime.fromisoformat("2020-01-01T10:00:01"),
    }
    record_b = {
        "id": 1,
        "value": "b",
        "_as_of": datetime.fromisoformat("2020-01-01T10:01:00"),
    }
    merged = merge_and_retain_last(
        spark_session.createDataFrame([record_b], record_schema),
        spark_session.createDataFrame([record_a1, record_a2], record_schema),
        key_fields=["id"],
        sort_field="_as_of",
        deleted_field="_deleted",
    )
    assert merged == DataFrameMatcher([record_b, record_a2])



from glue_biscuit.test.data_frame_matcher import DataFrameMatcher
from glue_biscuit.test.dynamic_frame_matcher import DynamicFrameMatcher
from glue_biscuit.test.fixtures import mock_glue_context, spark_context



import pytest
from pyspark.sql import SparkSession, DataFrame
from awsglue.context import GlueContext
from unittest.mock import patch, Mock


@pytest.fixture(scope="session")
def spark_context():
    spark = SparkSession.builder.master("local[1]").getOrCreate()
    yield spark.sparkContext


@pytest.fixture
def mock_glue_context(spark_context):
    mocked_methods = [
        "create_dynamic_frame_from_catalog",
        "create_dynamic_frame_from_options",
        "create_sample_dynamic_frame_from_catalog",
        "create_sample_dynamic_frame_from_options",
        "getSource",
        "create_data_frame_from_catalog",
        "create_data_frame_from_options",
        "forEachBatch",
        "getSink",
        "write_dynamic_frame_from_options",
        "write_from_options",
        "write_dynamic_frame_from_catalog",
        "write_dynamic_frame_from_jdbc_conf",
        "write_from_jdbc_conf",
        "purge_table",
        "purge_s3_path",
    ]
    gc = GlueContext(spark_context)
    context_mock = Mock(spec=gc, wraps=gc)

    for name in mocked_methods:
        method_mock: Mock = getattr(context_mock, name, None)
        if method_mock is not None:
            method_mock.return_value = None

    yield context_mock



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['glue_biscuit', 'glue_biscuit.test']

package_data = \
{'': ['*']}

install_requires = \
['black==22.3.0',
 'boto3==1.21.29',
 'botocore==1.24.46',
 'cfn-flip==1.3.0',
 'click==8.1.3',
 'jmespath==1.0.1',
 'mypy-extensions==0.4.3',
 'pathspec==0.10.1',
 'platformdirs==2.5.2',
 'python-dateutil==2.8.2',
 's3transfer==0.5.2',
 'six==1.16.0',
 'tomli==2.0.1',
 'typing-extensions==4.3.0',
 'urllib3==1.26.12']

setup_kwargs = {
    'name': 'glue-biscuit',
    'version': '0.0.1',
    'description': '',
    'long_description': '### Setup for local development (installing dependencies for IDE highlighting)\n\n1. Clone Repo\n1. Install Python 3.7\n   - Newer versions of Python 3 will probably work, but won\'t match the PyGlue\n1. Download [PyGlue.zip](https://s3.amazonaws.com/aws-glue-jes-prod-us-east-1-assets/etl-1.0/python/PyGlue.zip)\n   - Documentation around this can be found [here](https://docs.aws.amazon.com/glue/latest/dg/dev-endpoint-tutorial-pycharm.html)\n   - Put this zip in the project directory after cloning locally\n1. Install dependencies `pip install -r requirements.txt`\n1. (If using VSCode) Create `.env` in the project directory\n   - Add a line adding the PyGlue.zip file to your PYTHONPATH environment variable. e.g. `PYTHONPATH=".:/Users/bskiff/projects/python-stuff/GlueTest/PyGlue.zip"`\n\n### Running tests\n\n1. Install Docker\n1. Run `docker build -t glue . && docker run -it glue`\n\n### Linting and formatting\n\n```bash\nmake lint # checks linting and styles\nmake format # fixes style rule violations\n```\n',
    'author': 'Gene Tinderholm',
    'author_email': 'gtinderholm@sourceallies.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
